#include <allegro.h>
#include <string.h>
#include <stdio.h>

#include "globals.h"
#include "gui.h"

#include "poll.h"
#include "parport.h"
#include "simulate.h"


GUI::GUI() {
 
  lastpolls=0;
  numpolls=0;
  lastsecond=0.0;

  curfilename=0;

  graphtimestampdata=new double[GRAPHDATASIZE];
  graphdistancedata=new double[GRAPHDATASIZE];
  graphacceldata=new double[GRAPHDATASIZE];
  graphacceldatasmoothened=new double[GRAPHDATASIZE];
  graphpowerdata=new double[GRAPHDATASIZE];
  graphpowerdatasmoothened=new double[GRAPHDATASIZE];
  graphspeeddata=new double[GRAPHDATASIZE];
  graphspeeddatasmoothened=new double[GRAPHDATASIZE];
  graphrpmdata=new double[GRAPHDATASIZE];
  resetGraphData();

  generatedgraphx=new double[GRAPHDATASIZE];
  generatedgraphy=new double[GRAPHDATASIZE];

  doublebuffer=create_bitmap(RESOLUTIONX,RESOLUTIONY);
  clear_bitmap(doublebuffer);

  loadGraphics();
  prepareTachometer();
  prepareSpeedometer();
  prepareGraph();

  lastupdate=0.0;
  currentscreen=0;

  userinput=new char[10];
  userinput[0]='\0';
  menukeyboardcontrol=true;

}

GUI::~GUI() {

  destroy_bitmap(logo);
  destroy_bitmap(optionmenu);
  destroy_bitmap(carwireframe);
  for (int i=0;i<10;i++) {    
    destroy_bitmap(numbers[i]);
  }
  for (int i=0;i<13;i++) {    
    destroy_bitmap(numbers20[i]);
  }
  destroy_bitmap(needle);
  destroy_bitmap(whiteline);
  destroy_bitmap(smallredline);
  destroy_bitmap(bigredline);
  destroy_bitmap(rpmx1000);
  destroy_bitmap(kmperh);
  destroy_bitmap(bgcarimg);
  destroy_bitmap(powerhplabel);
  destroy_bitmap(enginespeedlabel);
  destroy_bitmap(enginespeedxlabel);
  destroy_bitmap(poweroutputhp);
  destroy_bitmap(time);
  destroy_bitmap(accelkmhpersec);
  destroy_bitmap(speedkmh);
 
  destroy_bitmap(doublebuffer);

}

void GUI::loadGraphics() {

  char *eachfilename;

  logo=load_pcx("images/uprlogo.pcx",NULL); 
  carwireframe=load_pcx("images/carbg.pcx",NULL);
  optionmenu=load_pcx("images/optmenu.pcx",NULL);
  for (int i=0;i<10;i++) {
    sprintf(eachfilename,"images/%i.pcx",i);
    numbers[i]=load_pcx(eachfilename,NULL);
  }
  for (int i=0;i<13;i++) {
    sprintf(eachfilename,"images/s%i.pcx",i*20);
    numbers20[i]=load_pcx(eachfilename,NULL);
  }
  needle=load_pcx("images/needle.pcx",NULL);
  whiteline=load_pcx("images/whiteln.pcx",NULL);
  smallredline=load_pcx("images/smredln.pcx",NULL);
  bigredline=load_pcx("images/bigredln.pcx",NULL);
  rpmx1000=load_pcx("images/rpmx1000.pcx",NULL);
  kmperh=load_pcx("images/kmh.pcx",NULL);
  bgcarimg=load_pcx("images/bgcarimg.pcx",NULL);
  powerhplabel=load_pcx("images/powerhp.pcx",NULL);
  enginespeedlabel=load_pcx("images/engspd.pcx",NULL);
  enginespeedxlabel=load_pcx("images/engrpm.pcx",NULL);
  poweroutputhp=load_pcx("images/powout.pcx",NULL);
  time=load_pcx("images/time.pcx",NULL);
  accelkmhpersec=load_pcx("images/accelkmh.pcx",NULL);
  speedkmh=load_pcx("images/speedkmh.pcx",NULL);

}

void GUI::prepareTachometer() {

  float numx;
  float numy;
  int count=0;
  int eachnum=0;

  //prepare image of tachometer
  //create it at one resolution (scale it as necessary for others)
  BITMAP *tachometerbig=create_bitmap(GAUGERESOLUTIONX,GAUGERESOLUTIONY);
  tachometer=create_bitmap(RESOLUTIONX*3/4,RESOLUTIONY*7/8);
  clear_bitmap(tachometerbig);
  clear_bitmap(tachometer);
  for (float i=-45.0;i<=225.0;i+=270.0*(200.0/MAXTACHOMETERRANGE)) {
    count=count%5;
    if (count==0) {
      numx=cos((i*256.0/360.0)+128.0)*170.0;
      numy=sin((i*256.0/360.0)+128.0)*170.0;
      draw_sprite(tachometerbig,numbers[eachnum],GAUGERESOLUTIONX/2+(int)numx-12,GAUGERESOLUTIONY/2+(int)numy-12);
      eachnum++;
    }

    if ((i+45.0)/270.0>TACHOMETERREDLINE/MAXTACHOMETERRANGE)
      pivot_sprite(tachometerbig,bigredline,GAUGERESOLUTIONX/2,GAUGERESOLUTIONY/2,150,5,ftofix(i*256.0/360.0));
    else {      
      if (count==0) {
        pivot_sprite(tachometerbig,whiteline,GAUGERESOLUTIONX/2,GAUGERESOLUTIONY/2,150,5,ftofix(i*256.0/360.0));                
      }
      else
        pivot_sprite(tachometerbig,smallredline,GAUGERESOLUTIONX/2,GAUGERESOLUTIONY/2,150,5,ftofix(i*256.0/360.0));      
    }

    count++;
  }
  draw_sprite(tachometerbig,rpmx1000,GAUGERESOLUTIONX/2-35,GAUGERESOLUTIONY/2+100);

  //scale tach to proper resolution
  stretch_sprite(tachometer,tachometerbig,0,0,RESOLUTIONX*3/4,RESOLUTIONY);

}

void GUI::prepareSpeedometer() {

  float numx;
  float numy;
  int count=0;
  int eachnum=0;

  //prepare image of speedometer
  //create it at one resolution (scale it as necessary for others)
  BITMAP *speedometerbig=create_bitmap(GAUGERESOLUTIONX,GAUGERESOLUTIONY);
  speedometer=create_bitmap(RESOLUTIONX*3/4,RESOLUTIONY*7/8);
  clear_bitmap(speedometerbig);
  clear_bitmap(speedometer);
  for (float i=-45.0;i<=225.0;i+=270.0*(2.0/MAXSPEEDOMETERRANGE)) {
    count=count%10;
    if (count==0) {
      numx=cos((i*256.0/360.0)+128.0)*170.0;
      numy=sin((i*256.0/360.0)+128.0)*170.0;
      draw_sprite(speedometerbig,numbers20[eachnum],GAUGERESOLUTIONX/2+(int)numx-12,GAUGERESOLUTIONY/2+(int)numy-12);
      eachnum++;
    }
   
    if (count==0) {
      pivot_sprite(speedometerbig,whiteline,GAUGERESOLUTIONX/2,GAUGERESOLUTIONY/2,150,5,ftofix(i*256.0/360.0));                
    }
    else if (count==5) {
      pivot_sprite(speedometerbig,bigredline,GAUGERESOLUTIONX/2,GAUGERESOLUTIONY/2,150,5,ftofix(i*256.0/360.0));                
    }
    else {
      pivot_sprite(speedometerbig,smallredline,GAUGERESOLUTIONX/2,GAUGERESOLUTIONY/2,150,5,ftofix(i*256.0/360.0));      
    }

    count++;
  }
  draw_sprite(speedometerbig,kmperh,GAUGERESOLUTIONX/2-25,GAUGERESOLUTIONY/2+100);

  //scale speedo to proper resolution
  stretch_sprite(speedometer,speedometerbig,0,0,RESOLUTIONX*3/4,RESOLUTIONY);

}

void GUI::prepareGraph() {
  
  GRAPHXSIZEPIXELS=RESOLUTIONX*3/4;
  GRAPHYSIZEPIXELS=RESOLUTIONY*3/4;
  graphbmp=create_bitmap(GRAPHXSIZEPIXELS+1,GRAPHYSIZEPIXELS+1);
 
  clear_bitmap(graphbmp);

  for (int i=1;i<=GRAPHXLINES;i++) {
    vline(graphbmp,i*GRAPHXSIZEPIXELS/GRAPHXLINES,0,GRAPHYSIZEPIXELS,4);     
  }

  for (int i=0;i<GRAPHYLINES;i++) {
    hline(graphbmp,0,i*GRAPHYSIZEPIXELS/GRAPHYLINES,GRAPHXSIZEPIXELS,4);
  }

  rect(graphbmp,0,0,GRAPHXSIZEPIXELS,GRAPHYSIZEPIXELS,15);
}

void GUI::resetGraphData() {

  for (int i=0;i<GRAPHDATASIZE;i++) {
    graphtimestampdata[i]=0.0;
    graphrpmdata[i]=0.0;
    graphspeeddata[i]=0.0;
    graphspeeddatasmoothened[i]=0.0;
    graphpowerdata[i]=0.0;
    graphacceldata[i]=0.0;
    
  }

  graphdataindex=0;

}

void GUI::resetGeneratedGraphData() {

  for (int i=0;i<GRAPHDATASIZE;i++) {
    generatedgraphx[i]=0.0;
    generatedgraphy[i]=0.0;
  }

}

void GUI::update(ParallelPortIO *ppio) {
   
  char *savefilename=new char[80];
  BITMAP *bmp;
  PALETTE pal;

  int keyread;

  //update only UPDATESPERSEC times per second
  if (((double)clock())/CLOCKS_PER_SEC-lastupdate<=1.0/UPDATESPERSEC) return;
  lastupdate=((double)clock())/CLOCKS_PER_SEC;

  clear_to_color(doublebuffer, 16);

  //background graphic updates
  draw_sprite(doublebuffer,logo,0,0);
  draw_sprite(doublebuffer,carwireframe,RESOLUTIONX-carwireframe->w,RESOLUTIONY-carwireframe->h);

  //recalculate data
  calculateGraphData(ppio);

  if (currentscreen==0) {  //main menu
    draw_sprite(doublebuffer,optionmenu,0,RESOLUTIONY/8);
    textprintf(doublebuffer, font, RESOLUTIONX/2,RESOLUTIONY/4,15, "Trip: %.3f", TRIPOMETER);
  }
  else if (currentscreen==1) //tachometer
    drawTachometer();
  else if (currentscreen==2)
    drawSpeedometer();
  else if (currentscreen==3)
    drawPowerOverTimeGraph();
  else if (currentscreen==4)
    drawVelocityOverTimeGraph();
  else if (currentscreen==5)
    drawAccelerationOverTimeGraph();
  else if (currentscreen==6)
    conductDynamometerRun();
  else if (currentscreen==7) {
    if (currentscreenstate==0) {
      if (obtainString("Enter the filename of the dynamo graph to load:",
           RESOLUTIONX/8,RESOLUTIONY/8)) {
        displayMessage("Loading...",0,0);        
        if (loadDynamoGraph())
          currentscreenstate=1;
        else {
          currentscreen=0;
          displayMessage("Unable to load file!",RESOLUTIONX/8,RESOLUTIONY/8);
        }
      }
    }
    else if (currentscreenstate==1)
      drawGeneratedDynamoGraph();
  }

  draw_sprite(screen,doublebuffer,0,0);

  //handle keypresses
  if (keypressed()&&menukeyboardcontrol) {
    keyread=readkey();
    keyread=keyread&0xff;

    if (keyread=='q') {      
      exit(0);    
    }
    else if (keyread=='1')
      currentscreen=1;
    else if (keyread=='2')
      currentscreen=2;
    else if (keyread=='3') {
      currentscreen=3;
    }
    else if (keyread=='4') {
      currentscreen=4;
    }
    else if (keyread=='5') {
      currentscreen=5;      
    }
    else if (keyread=='6') {
      currentscreen=6;
      currentscreenstate=0;
    }
    else if (keyread=='7') {
      currentscreen=7;
      currentscreenstate=0;      
    }
    else if (keyread=='s'||keyread=='S') { //save bitmap
      sprintf(savefilename,"scrn%i.bmp",curfilename);
      bmp = create_sub_bitmap(screen,0,0,RESOLUTIONX,RESOLUTIONY);
      get_palette(pal);
      save_bitmap(savefilename,bmp,pal);
      curfilename++;
      destroy_bitmap(bmp);
    }
    else if (keyread=='e'||keyread=='E') {
      if (currentscreen==6&&currentscreenstate==3) {   
        currentscreenstate=4;
      }
    }
    else
      currentscreen=0;
  }  

  if (menukeyboardcontrol) clear_keybuf();

}

void GUI::incrementPoll() {
  numpolls++;
}

double GUI::calculateGraphData(ParallelPortIO *ppio) {
  
  graphdataindex=(graphdataindex+1)%GRAPHDATASIZE;

  graphdistancedata[graphdataindex]=TRIPOMETER;
 
  graphtimestampdata[graphdataindex]=(double(clock())/CLOCKS_PER_SEC);

  graphrpmdata[graphdataindex]=calculateTachometer(ppio->obtainPolls(TACHBUSADDR));

  graphspeeddata[graphdataindex]=calculateSpeedometer(ppio->obtainPolls(SPEEDBUSADDR));

  averageSmooth(graphspeeddata,graphspeeddatasmoothened,AVGSMOOTHSIZE);

  graphacceldata[graphdataindex]=calculateAccelerometer();
  if (graphacceldata[graphdataindex]>ACCELERATIONOVERTIMERANGE_KMHSEC)
    graphacceldata[graphdataindex]=ACCELERATIONOVERTIMERANGE_KMHSEC;
  else if (graphacceldata[graphdataindex]<0.0)
    graphacceldata[graphdataindex]=0.0;

  averageSmooth(graphacceldata,graphacceldatasmoothened,AVGSMOOTHSIZE);

  graphpowerdata[graphdataindex]=calculatePowerOutput();
  if (graphpowerdata[graphdataindex]>POWEROVERTIMECURVERANGE_HP)
    graphpowerdata[graphdataindex]=POWEROVERTIMECURVERANGE_HP;
  else if (graphpowerdata[graphdataindex]<0.0)
    graphpowerdata[graphdataindex]=0.0;

  averageSmooth(graphpowerdata,graphpowerdatasmoothened,AVGSMOOTHSIZE);

}

double GUI::calculateTachometer(Poll *sequence) {

  //calculated by considering the number of pulses to occur in a second
  //also note that 2 pulses occur for each revolution
  //preliminary formula:
  // E(pollreadingsinlastsecond)*60(sec/min)/2(pulse/revolution)

  double calculatedrpm=0.0;
  int count=0;  

  for (int i=0;i<NUMINPUTSAMPLES;i++) {
    if ((((double)clock())/CLOCKS_PER_SEC-sequence[i].getTimestamp())<PULSETIMECONSIDERED&&sequence[i].getReading()<=15) {
      calculatedrpm+=(double)(sequence[i].getReading());
      count++;
    }
  }

  calculatedrpm=(calculatedrpm*60.0)/(2.0*PULSETIMECONSIDERED);
  return calculatedrpm;
}

double GUI::calculateSpeedometer(Poll *sequence) {

  double calculatedspeed=0.0;
  int count=0;

  for (int i=0;i<NUMINPUTSAMPLES;i++) {
    if ((((double)clock())/CLOCKS_PER_SEC-sequence[i].getTimestamp())<PULSETIMECONSIDERED&&sequence[i].getReading()<=15) {
      calculatedspeed+=(double)(sequence[i].getReading());      
      count++;
    }    
  }   

  calculatedspeed=(calculatedspeed*3600.0)/(PULSETIMECONSIDERED*SPEEDPULSESPERKM);
  return calculatedspeed;

}

double GUI::interpolate(double *data, double howlongago) {

  //determine upper and lower bound known values for interpolate
  double INTERPOLATETIME=graphtimestampdata[graphdataindex]-howlongago;
  int lowerindex=-1;
  for (int i=0;i<GRAPHDATASIZE;i++) {
    if ((graphtimestampdata[i]>graphtimestampdata[lowerindex]&&graphtimestampdata[i]<INTERPOLATETIME)||
        (graphtimestampdata[i]<INTERPOLATETIME&&graphtimestampdata[lowerindex]>INTERPOLATETIME))
      lowerindex=i;
  }

  int upperindex=-1;
  for (int i=0;i<GRAPHDATASIZE;i++) {
    if ((graphtimestampdata[i]<graphtimestampdata[upperindex]&&graphtimestampdata[i]>INTERPOLATETIME)||
        (graphtimestampdata[i]>INTERPOLATETIME&&graphtimestampdata[upperindex]<INTERPOLATETIME))
      upperindex=i;
  }

  //interpolate a velocity from the bounding known velocities
  double interpolatedomain=graphtimestampdata[lowerindex]-graphtimestampdata[upperindex];  
  double value=((graphtimestampdata[lowerindex]-INTERPOLATETIME)/interpolatedomain)*data[lowerindex]+
                           ((INTERPOLATETIME-graphtimestampdata[upperindex])/interpolatedomain)*data[upperindex];

  return value;

}

void GUI::averageSmooth(double *srcdata, double *destdata, int width) {

  //graphdataindex

  double total;
  int counted;
  int index;
  bool wraparound;

  for (int i=graphdataindex+1;i<GRAPHDATASIZE;i++) {
    total=0.0;
    counted=0;

    for (int j=i-width;j<=i;j++) {
      wraparound=false;
 
      index=j;
      if (index>=GRAPHDATASIZE) {
        index=index%GRAPHDATASIZE;
        wraparound=true;
      }

      if ((!wraparound&&index>graphdataindex)||
          (wraparound&&index<=graphdataindex)) {
        total+=srcdata[index];
        counted++;
      }

    }

    if (counted>0)
      destdata[i]=total/(double)counted;
    else
      destdata[i]=0.0;

  }

  for (int i=0;i<=graphdataindex;i++) {
     
    total=0.0;
    counted=0;

    for (int j=i-width;j<=i;j++) {
      wraparound=true;
 
      index=j;
      if (index<0) {
        index=index+GRAPHDATASIZE;
        wraparound=false;
      }

      if ((!wraparound&&index>graphdataindex)||
          (wraparound&&index<=graphdataindex)) {
        total+=srcdata[index];
        counted++;
      }

    }

    if (counted>0)
      destdata[i]=total/(double)counted;
    else
      destdata[i]=0.0;
  }

}

double GUI::calculateAccelerometer() {

  //acceleration is delta velocity (the difference in velocities)
  double acceleration=(graphspeeddatasmoothened[graphdataindex]-interpolate(graphspeeddatasmoothened,CALC_H_APPROX));  
  acceleration=acceleration/CALC_H_APPROX;  
 
  return acceleration; //value returned is in (km/h)/s

}

double GUI::calculatePowerOutput() {

/*
Formulas:
F=ma
Energy=F*dist
PowerOutput=Energy/Duration

dist is approximated by taking the midpoint between velocities at two times, and multiplying it
by the difference in time the two velocities were taken at.

graphdataindex here refers to the index into the graphdata of the current polled info

*/
  
  //F=ma
  //since the acceleration given is in km/h - convert to m/s   
  //F is in Newtons, kg*m/s^2
  
  double mass=VEHICLEWEIGHTKG;              
  double acceleration=graphacceldatasmoothened[graphdataindex]*1000.0/3600.0;  
  double forceapplied=mass*acceleration;
  
  //Energy=F*dist
  //the dist value (the distance travelled in the last second)
  //convert from km to m
      
  double dist=(graphdistancedata[graphdataindex]-interpolate(graphdistancedata,1.0))*1000.0;    
  double energy=forceapplied*dist; 
 
  //PowerOutput=Work/Duration (the unit will be Joules/sec, or Watts)
  //since the duration taken was for 1 second, the formula is simply power=energy/1.0
  double poweroutput=energy;

  //Conversion from Watt to HP: 1HP = 746 Watts (approximately)
  //return poweroutput/746.0;  
  return poweroutput/746.0;
}

void GUI::drawTachometer() {

  //"fixed" is an allegro-specific unit of measure for an angle
  //a 32-bit data structure that represents a fixed point number
  //(a complete circle in "fixed" is represented by 256)
  
  double calculatedrpm;
  float needleangle;

  //draw tachometer background sprite at correct scale
  draw_sprite(doublebuffer,tachometer,RESOLUTIONX/8,RESOLUTIONY/8);

  //process polls and calculate the rpm
  calculatedrpm=graphrpmdata[graphdataindex];
    
  //draw needle
  needleangle=-45.0+(270.0*calculatedrpm/MAXTACHOMETERRANGE);
  pivot_scaled_sprite(doublebuffer,needle,RESOLUTIONX/2,RESOLUTIONY*5/8,101,7,ftofix(needleangle*256.0/360.0),ftofix((float)RESOLUTIONY/GAUGERESOLUTIONY));
  
  textprintf(doublebuffer, font, 0,RESOLUTIONY/4,15, "RPM: %.1f", calculatedrpm);
  textprintf(doublebuffer, font, 0,RESOLUTIONY/4+10,15, "PPS: %i", lastpolls);

  if (((double)clock())/CLOCKS_PER_SEC>lastsecond+1.0) {
    lastpolls=numpolls;
    numpolls=0;
    lastsecond=((double)clock())/CLOCKS_PER_SEC;
  }

}

void GUI::drawSpeedometer() {
 
  double calculatedspeed;
  float needleangle;

  //draw tachometer sprite at correct scale
  draw_sprite(doublebuffer,speedometer,RESOLUTIONX/8,RESOLUTIONY/8);

  //process polls and calculate speed
  calculatedspeed=graphspeeddatasmoothened[graphdataindex];  
    
  //draw needle
  needleangle=-45.0+(270.0*calculatedspeed/MAXSPEEDOMETERRANGE);
  pivot_scaled_sprite(doublebuffer,needle,RESOLUTIONX/2,RESOLUTIONY*5/8,101,7,ftofix(needleangle*256.0/360.0),ftofix((float)RESOLUTIONY/GAUGERESOLUTIONY));
  
  textprintf(doublebuffer, font, 0,RESOLUTIONY/4,15, "SPD: %.1f", calculatedspeed);
  textprintf(doublebuffer, font, 0,RESOLUTIONY/4+10,15, "PPS: %i", lastpolls);

  if ((double)clock()/CLOCKS_PER_SEC>lastsecond+1.0) {
    lastpolls=numpolls;
    numpolls=0;
    lastsecond=(double)clock()/CLOCKS_PER_SEC;
  }

}

void GUI::drawPowerOverTimeGraph() {

  drawGraph(graphtimestampdata, graphpowerdatasmoothened,
            time,poweroutputhp,false, ((double)clock())/CLOCKS_PER_SEC-POWEROVERTIMECURVEDURATION_SEC,POWEROVERTIMECURVEDURATION_SEC,1,0,POWEROVERTIMECURVERANGE_HP,10);
  
}

void GUI::drawAccelerationOverTimeGraph() {

  drawGraph(graphtimestampdata, graphacceldatasmoothened, 
            time,accelkmhpersec,false, ((double)clock())/CLOCKS_PER_SEC-ACCELERATIONOVERTIMEDURATION_SEC,ACCELERATIONOVERTIMEDURATION_SEC,1,0,ACCELERATIONOVERTIMERANGE_KMHSEC,2);

}

void GUI::drawVelocityOverTimeGraph() { 

  drawGraph(graphtimestampdata, graphspeeddatasmoothened,
            time,speedkmh,false, ((double)clock())/CLOCKS_PER_SEC-VELOCITYOVERTIMEDURATION_SEC,VELOCITYOVERTIMEDURATION_SEC,1,0,MAXSPEEDOMETERRANGE,10);

}

void GUI::drawPowerOverRPMGraph() {

  drawGraph(graphrpmdata, graphpowerdatasmoothened, enginespeedxlabel, poweroutputhp,
    false,1500,TACHOMETERREDLINE-1500,1000,0,POWEROVERTIMECURVERANGE_HP,10);  
  
}

void GUI::drawGeneratedDynamoGraph() {

  drawGraph(generatedgraphx, generatedgraphy, enginespeedxlabel, poweroutputhp,
    false,1500,TACHOMETERREDLINE-1500,1000,0,POWEROVERTIMECURVERANGE_HP,10);  

}

void GUI::drawGraph(double *xdata, double *ydata, BITMAP *xaxislabel, BITMAP *yaxislabel, bool XAXISFLIP, double XAXISSTART, double XAXISSIZE, int XAXISINTERVAL, double YAXISSTART, double YAXISSIZE, int YAXISINTERVAL) {

  draw_sprite(doublebuffer,graphbmp,RESOLUTIONX/8,RESOLUTIONY/8);  
 
  //textprintf(doublebuffer, font, 0,RESOLUTIONY/8-20,15, "Index: %i", graphdataindex);   
  textprintf(doublebuffer, font, 0,RESOLUTIONY/8-20,15, "YData: %.1f", ydata[graphdataindex]);  

  //graph data acts as a circular array.
  //display lines appropriately
  
  double startx, starty; //origin is bottom left corner of graph
  double endx, endy; //origin is bottom left corner of graph
  int eachdata;
  bool finished=false;
 
  if (XAXISFLIP)
    startx=GRAPHXSIZEPIXELS-((xdata[graphdataindex]-XAXISSTART)/XAXISSIZE)*GRAPHXSIZEPIXELS;
  else
    startx=((xdata[graphdataindex]-XAXISSTART)/XAXISSIZE)*GRAPHXSIZEPIXELS;
  starty=((ydata[graphdataindex]-YAXISSTART)/YAXISSIZE)*GRAPHYSIZEPIXELS;

  eachdata=(graphdataindex-1+GRAPHDATASIZE)%GRAPHDATASIZE;
  do {
    
    if (XAXISFLIP)
      endx=GRAPHXSIZEPIXELS-((xdata[eachdata]-XAXISSTART)/XAXISSIZE)*GRAPHXSIZEPIXELS;
    else
      endx=((xdata[eachdata]-XAXISSTART)/XAXISSIZE)*GRAPHXSIZEPIXELS;
    endy=((ydata[eachdata]-YAXISSTART)/YAXISSIZE)*GRAPHYSIZEPIXELS;
    
    if (startx>=0&&startx<=GRAPHXSIZEPIXELS
        &&starty>=0&&starty<=GRAPHYSIZEPIXELS
        &&endx>=0&&endx<=GRAPHXSIZEPIXELS
        &&endy>=0&&endy<=GRAPHYSIZEPIXELS) 
      line(doublebuffer,RESOLUTIONX/8+(int)startx,RESOLUTIONY/8+GRAPHYSIZEPIXELS-(int)starty,RESOLUTIONX/8+(int)endx,RESOLUTIONY/8+GRAPHYSIZEPIXELS-(int)endy,15);      

    startx=endx;
    starty=endy;    

    eachdata=(eachdata-1+GRAPHDATASIZE)%GRAPHDATASIZE;
    if (eachdata==graphdataindex)
      finished=true;

  } while (!finished);
  
  draw_sprite(doublebuffer,yaxislabel,RESOLUTIONX/8-24-yaxislabel->w,RESOLUTIONY/2-(yaxislabel->h/2));
  draw_sprite(doublebuffer,xaxislabel,RESOLUTIONX/2-((xaxislabel->w)/2),RESOLUTIONY/8+GRAPHYSIZEPIXELS+15);

  //draw axis unit labels
  for (int i=0;i<=YAXISSIZE;i+=YAXISINTERVAL) {
    textprintf(doublebuffer, font,
               RESOLUTIONX/8-25,
               (int)(RESOLUTIONY/8+GRAPHYSIZEPIXELS-(((double)i/YAXISSIZE)*GRAPHYSIZEPIXELS)-4),
               15, "%.0f", ((((double)i)/YAXISSIZE)*YAXISSIZE)+YAXISSTART);   
  }  
  for (int i=0;i<=XAXISSIZE;i+=XAXISINTERVAL) {
    if (XAXISFLIP)
      textprintf(doublebuffer, font,
               (int)(RESOLUTIONX/8+GRAPHXSIZEPIXELS-(((double)i/XAXISSIZE)*GRAPHXSIZEPIXELS)-12),
               RESOLUTIONY/8+GRAPHYSIZEPIXELS+4,
               15, "%.0f", ((((double)i)/XAXISSIZE)*XAXISSIZE)+XAXISSTART);   
    else
      textprintf(doublebuffer, font,
               (int)(RESOLUTIONX/8+(((double)i/XAXISSIZE)*GRAPHXSIZEPIXELS)-12),
               RESOLUTIONY/8+GRAPHYSIZEPIXELS+4,
               15, "%.0f", ((((double)i)/XAXISSIZE)*XAXISSIZE)+XAXISSTART);   
  }

}

void GUI::conductDynamometerRun() {

  //dyno run occurs through FSM states
  //state 0 -> vehicle idling
  //state 1 -> acceleration (and recording of data)
  //state 2 -> brief deceleration (to calculate drag forces)
  //state 3 -> presentation of data (with optional saving)
  if (currentscreenstate==0) { //slow vehicle

     drawTachometer();     

     displayMessage("Lower Engine RPM to below 1500.",RESOLUTIONX/8,RESOLUTIONY/8);     

     if (graphrpmdata[graphdataindex]<1500)
       currentscreenstate++;
  }
  else if (currentscreenstate==1) { //instruct to drive to redline
     
     drawTachometer();

     displayMessage("Accelerate vehicle to Engine RPM Redline.",RESOLUTIONX/8,RESOLUTIONY/8);

     if (graphrpmdata[graphdataindex]>=1500) {
       currentscreenstate++;            
       resetGeneratedGraphData();       
     }
  }
  else if (currentscreenstate==2) { //record data
     
    displayMessage("Recording...",RESOLUTIONX/4,RESOLUTIONY/8);
    
    drawTachometer();     

    //copy data over to generated graph
    generatedgraphx[graphdataindex]=graphrpmdata[graphdataindex];
    generatedgraphy[graphdataindex]=graphpowerdata[graphdataindex];

    //when engine RPM's begin to exceed range, complete recording
    if (graphrpmdata[graphdataindex]>=TACHOMETERREDLINE-250) {
      currentscreenstate=3;
    }
  }
  else if (currentscreenstate==3) { //display generated dynamometer curve
    drawGeneratedDynamoGraph();

    displayMessage("E - Export Data to File",RESOLUTIONX/4,RESOLUTIONY-20);
    
  }
  else if (currentscreenstate==4) { //user wishes to save

    drawGeneratedDynamoGraph();
     
    if (obtainString("Enter the filename to save graph as:",RESOLUTIONX/8,RESOLUTIONY/8)) {
      displayMessage("Saving...",0,0);
      saveDynamoGraph();
      currentscreenstate=3;
    }

  }

}

void GUI::saveDynamoGraph() {

  char *filenameandpath=new char[80];
  FILE *f;

  sprintf(filenameandpath,"graphs/%s",userinput);
  f=fopen(filenameandpath,"w");
  if (f!=NULL) {
    fprintf(f,"DYNAMOMETERGRAPH\n");
    for (int i=0;i<GRAPHDATASIZE;i++) {
      fprintf(f,"%.3f %.3f\n",generatedgraphx[i],generatedgraphy[i]);
    }
    displayMessage("Saved.",0,0);
  }
  else {
    displayMessage("Unable to save!",0,0);
  }

  fclose(f);

}

bool GUI::loadDynamoGraph() {

  char *filenameandpath=new char[80];
  char *eachline=new char[80];
  char *xvalue=new char[40];
  char *yvalue=new char[40];
  FILE *f;
  int graphindex;

  sprintf(filenameandpath,"graphs/%s",userinput);
  f=fopen(filenameandpath,"r");

  if (f==NULL) {
    displayMessage("File not found!",0,0);    
    return false;
  }

  fgets(eachline,80,f);
 
  if (strncmp(eachline,"DYNAMOMETERGRAPH",16)!=0) {
    displayMessage("Invalid file format!",0,0);
    return false;
  }

  //reset all data before copy
  resetGeneratedGraphData();
  graphindex=0;

  while (!feof(f)) {
    
    fgets(eachline,80,f);

    if (strlen(eachline)>2) {
      strncpy(xvalue,eachline,(strchr(eachline,' ')-eachline));
      xvalue[strchr(eachline,' ')-eachline]='\0';
      strcpy(yvalue,strchr(eachline,' ')+1);     

      generatedgraphx[graphindex]=atof(xvalue);
      generatedgraphy[graphindex]=atof(yvalue);

    }

    graphindex=(graphindex+1)%GRAPHDATASIZE;

  }

  fclose(f);

}

void GUI::displayMessage(char *message, int xlocation, int ylocation) {

  rectfill(doublebuffer,xlocation,ylocation,
    xlocation+(strlen(message)*8+10),
    ylocation+18,16);
  rect(doublebuffer,xlocation,ylocation,
    xlocation+(strlen(message)*8+10),
    ylocation+18,4);

  textprintf(doublebuffer, font,xlocation+5,ylocation+5,15, 
    message); 
 
}

bool GUI::obtainString(char *message, int xlocation, int ylocation) {

  menukeyboardcontrol=false;
  int keyread;
  char *inputchar=new char[2];
  inputchar[1]='\0';

  rectfill(doublebuffer,xlocation,ylocation,
    xlocation+(strlen(message)*8+10),
    ylocation+26,16);
  rect(doublebuffer,xlocation,ylocation,
    xlocation+(strlen(message)*8+10),
    ylocation+26,4);

  textprintf(doublebuffer, font,xlocation+5,ylocation+5,15, 
    message);    

  textprintf(doublebuffer, font,xlocation+5,ylocation+15,15,
    userinput);  

  if (keypressed()) {
    keyread=readkey();
    keyread=keyread&0xff;

    if (((keyread>=65&&keyread<=90)||(keyread>=97&&keyread<=122)||(keyread>=48&&keyread<=57))
        &&strlen(userinput)<=7) { //alphanumerics
      inputchar[0]=keyread;
      strcat(userinput,inputchar);
    }
    else if (keyread==8&&strlen(userinput)>0) { //backspace
      userinput[strlen(userinput)-1]='\0';
    }
    else if (keyread==13&&strlen(userinput)>0) {  //enter
      //forfeit keyboard control for this input
      menukeyboardcontrol=true;
      return true;
    }

  }

  clear_keybuf();

  return false;

}


