#include <allegro.h>
#include <string.h>
#include <stdio.h>

#include "globals.h"

#include "poll.h"
#include "parport.h"


ParallelPortIO::ParallelPortIO() {
  
  beginindex=new int[NUMINPUTSENSORS];

  inputsequence=new Poll*[NUMINPUTSENSORS];  
  for (int i=0;i<NUMINPUTSENSORS;i++) {
    beginindex[i]=0;
    inputsequence[i]=new Poll[NUMINPUTSAMPLES];    
  }

}

void ParallelPortIO::setSimulated(bool simulated) {

  this->simulated=simulated;

}

bool ParallelPortIO::getSimulated() {

  return simulated;

}

/*
 * Note: To be used by Simulator only
 */
void ParallelPortIO::addPoll(int whichsensor, double timestamp, unsigned char reading) {

  //set reading and timestamp for poll
  inputsequence[whichsensor][beginindex[whichsensor]].setReading(reading);
  inputsequence[whichsensor][beginindex[whichsensor]].setTimestamp(timestamp);

  //iterate to next poll element in array
  beginindex[whichsensor]++;
  beginindex[whichsensor]=beginindex[whichsensor]%NUMINPUTSAMPLES;

}

int ParallelPortIO::pollSensor(int sensoraddress) {

  outp(outport, (1<<(sensoraddress*2)));

  //delay one-thousandth of a second to wait for the hardware to send
  //data to the bus

  uclock_t delaytime=uclock();
  while (uclock()-delaytime<UCLOCKS_PER_SEC*0.001) { }

  int busdata=inp(inport);
  outp(outport, (2<<(sensoraddress*2)));
  outp(outport, 0);

  return (busdata>>4)^8;

}

Poll *ParallelPortIO::obtainPolls(int busaddress) {

  return inputsequence[busaddress];

}

Poll ParallelPortIO::obtainLastPoll(int busaddress) {

  return inputsequence[busaddress][((beginindex[busaddress]+NUMINPUTSAMPLES-1)%NUMINPUTSAMPLES)];

}

